/* 
 Standard Bag of features code
 =============================
 Reads in a codebook file, a datafile. 
 performs quantization and creates a histogram in the form of a row vector. 
 Please refer to README for more details.
 subh@cs.ucf.edu 
*/
#include "codebook.h"
int main(int argc, char **argv)
{
  /* Input Data Related variables */
  P_PARAMS cp;
  float *data=NULL;
  long int nDataPts =0, nDataDim =0;
  /* Codebook related variables */
  float *centers_data=NULL, *data_dists=NULL;
  int *memberships=NULL;
  long int codebook_dim = 0, num_clusters = 0;
  /* Bag of words related variables - This is row vector */
  CvMat *bow_representation;
  FLANNParameters p;
  FLANN_INDEX index_id;
  float speedup;

  if (parse_options (argc, argv, &cp) !=0)
    return 1;
  
  /* Read single input data file */
  data = parse_data_file (cp.data_file, &nDataPts, &nDataDim);
  fprintf (stderr, "INF:\n" 
                   "  Total Number of datapoints : %ld\n"
                   "  Datapoint Dimension        : %ld\n", 
      nDataPts, nDataDim);
  
 /* allocating storage for cluster memberships and distances */
  data_dists = (float *)  malloc (nDataPts * sizeof(float));
  memberships = (int *) malloc (nDataPts * sizeof (int));
    
  /* Read in codebook file */
  centers_data = parse_data_file (cp.codebook_file, &num_clusters, &codebook_dim);
  cp.nclusters = num_clusters;
  init_FLANN_indexing (&p);
  
  /* If pre-built index is not available, then build index otherwise 
  load index from file */
  if( access(cp.idx_file, R_OK) == 0){
    fprintf (stderr, "INF:Index file %s present.Loading indices.\n", cp.idx_file);
    index_id = flann_load_index(cp.idx_file, centers_data, cp.nclusters, codebook_dim);
    fprintf (stderr, "INF:Index file %s successfully loaded.\n", cp.idx_file);
  } else {
    fprintf (stderr, "INF:Index file %s absent. Building indices.\n", cp.idx_file);
    index_id = flann_build_index (centers_data, cp.nclusters, codebook_dim, 
        &speedup, &p);
    if(flann_save_index(index_id, cp.idx_file)==0)
      fprintf (stderr, "INF:Index file %s saved.\n", cp.idx_file);
    else
      fprintf (stderr, "ERR:Unable to save index file %s.\n", cp.idx_file);
  }  
  /* Searching for nearest neighbors among codebook words for all datapoints in 
   the datafile */
  flann_find_nearest_neighbors_index (index_id, data, nDataPts, memberships, 
        data_dists, NN, &p); 
  
  /* Generate bag-of-words representation for datapoints */    
  bow_representation = cvCreateMat (1, cp.nclusters, CV_32FC1);
  
  fprintf (stderr, "INF:Quantizing Single File with %ld points, %ld dimensions\n", nDataPts, nDataDim);
  quantize_single_file (centers_data, data_dists, cp.nclusters, 
      memberships, nDataPts, &bow_representation);
  
  fprintf (stderr, "INF:Writing results into file\n");
  print_matrix (bow_representation, cp.bow_file);
  
  cvReleaseMat (&bow_representation);
  free (data);
  free (memberships);
  free (centers_data);
  free (data_dists);
  return 0;
}
// Implementation specific files 
void usage (char* pname) {
  fprintf (stderr, "Usage: %s\n"
     "  -d <datafile>         [Input query file with descriptors]\n"
     "  -b <codebook file>    [Input codebook file with cluster centers]\n"
     "  -o <output bow file>  [Output BoW file path]\n"
     "  -i <index file>       [Output index file path]\n", pname);
}
//Parse command line options for codebook creation program
int parse_options (int argc, char *argv[], P_PARAMS *p) {
  int i;
  
  if( argc == 1) {
    usage (argv[0]);
    return 1;
  }
  for (i = 1; i < argc; i++ ){
    if (!strcmp (argv[i], "-d")) /* Data file */
      strcpy (p->data_file , argv[++i]);
    else if (!strcmp (argv[i], "-b"))  /*code book file*/
      strcpy (p->codebook_file,  argv[++i]);
    else if (!strcmp (argv[i], "-o")) /* Output bag of words file */ 
      strcpy (p->bow_file,  argv[++i]);
    else if (!strcmp (argv[i], "-i")) /* Index file [if available] */
      strcpy (p->idx_file,  argv[++i]);
  }
  fprintf (stderr, "INF:Using parameters\n");
  fprintf (stderr, "  Input codebook file      : %s\n" 
                   "  Input Descriptors File   : %s\n"
                   "  Output Bag Of Words File : %s\n"
                   "  Index file [Created once]: %s\n", 
           p->codebook_file, p->data_file, p->bow_file, p->idx_file);
  return 0;
}

void quantize_single_file (float *centers_data, float *data_dists, int nclusters, 
    int *memberships,long int nDataPts, CvMat **bow_representation) {
  long int j;
  int k, cluster_idx;
  // required for normalizing the feature vectors
  CvMat *rowvec;

  //Vector to store histograms - this is our bag of words representation 
  cvSetZero (*bow_representation);
  
  for (j = 0; j < nDataPts; j++) {
    cluster_idx = memberships[j];
    // Creating a weighted histogram instead of a simple histogram
    //cvmSet (*bow_representation, 0, cluster_idx, 
    //    cvmGet(*bow_representation, 0, cluster_idx)+(float)1/(data_dists[j]+0.00001));
    cvmSet (*bow_representation, 0, cluster_idx, 
       cvmGet(*bow_representation, 0, cluster_idx)+1);

  }
  /*Normalize each row of the codebook (standard procedure to make sure we 
  dont bias the feature vector to a larger number of descriptors) */
  rowvec = cvCreateMat (1, nclusters, CV_32FC1);

  //Normalize the descriptor to a before putting it in the descriptor matrix
  for (k=0; k<nclusters; k++)
    cvmSet (rowvec, 0, k, cvmGet(*bow_representation, 0, k));
   cvNormalize (rowvec, rowvec, 0, 1, CV_MINMAX, NULL);

  for (k=0; k<nclusters; k++)
    cvmSet (*bow_representation, 0, k, cvmGet (rowvec, 0, k));
 
  cvReleaseMat (&rowvec);
}
void init_FLANN_clustering (FLANNParameters *p) {
  // FLANN clustering parameters
  p->algorithm = FLANN_INDEX_KMEANS;
  p->checks = 2048;
  p->cb_index = 0.6;
  //p->trees = 64; //(rec val)
  p->branching = 2;
  p->iterations = 1000;
  p->centers_init = FLANN_CENTERS_GONZALES;
  p->target_precision = -1;
  p->build_weight = 0.01;
  p->memory_weight = 1;
}
void init_FLANN_indexing (FLANNParameters* p) {
  p->algorithm = FLANN_INDEX_KDTREE;
  p->checks =2048;
  p->trees = 16;
  p->target_precision = -1;
  p->build_weight = 0.01;
  p->memory_weight = 1;
}
void init_FLANN_searching (FLANNParameters* p){
  p->log_level = FLANN_LOG_ERROR;
  p->random_seed = FLANN_CENTERS_RANDOM; 
}
